%% MATLAB (R) script to compute the MMSE-DFE slicer input eye diagram %%%%%%%%%%

%% Version: 1.1
%% Date: September 26, 2007
%% Author: Adam Healey, LSI Corporation

clear variables
warning off         % Suppress "divide-by-0" and "log of zero" warnings

%% Define inputs %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% sampleFile : Contains exactly "oversampling" samples per symbol. The samples
%% must be circularly shifted to align with the symbols in "symbolFile." The
%% file format is a single column of chronological numerical samples, in ASCII
%% format, with no header or footer
%% 
%% symbolFile : The file format is a single column of chronological symbols, in
%% ASCII format, with no header or footer
%% sampleFile      = 'samples.txt';    
%% symbolFile      = 'symbols.txt';
sampleFile      = 'SAS_CJTPAT_samples.txt';    
symbolFile      = 'SAS_CJTPAT_symbols.txt';
oversampling    = 16;           % samples/symbol, must be an even integer

%% Define constants %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Note that T is the symbol period (normalized to 1) and "eqNc" must be 1 or 
%% an even integer
ber0            = 1E-15;        % target bit error ratio
maxEqDelay      = 16;           % maximum equalizer delay (symbol periods)
eqNc            = 1;            % number of feed-forward taps (T/2-spaced)
eqNd            = 3;            % number of feedback taps (T-spaced)
dphi            = 1/100;        % eye diagram phase step (symbol periods)
dvee            = 1/100;        % eye diagram amplitude step

graphOut        = 1;            % control graphical outputs
                                % 0 : disable graphical outputs
                                % 1 : bit error ratio map

%% Define stressors %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% stressorFile : File format is a column of values, in ASCII format, for each
%% stressor, with no header or footer. Each column is to be separated by white
%% space. The first row of each column contains the RMS noise amplitude and is
%% followed by chronological numerical samples of impulse response sampled at 
%% T/"oversampling" intervals.
%% stressorFile    = '070913_BETA_EPSILON_RX.txt';
stressorFile    = 'DFEEYE_stressor_iPass_10m_h0Table_6g0_16X.txt';
%% stressorFile    = '070925_TCTF_hATable_16X.txt';
%% stressorFile    = '070925_TCTF_h0Table_16X.txt';

h0 = load( stressorFile );
rmsNoise = h0(1, :);
h0 = h0(2:end, :);

%% Load input waveform and symbol sequence %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
y0 = load( sampleFile );
x0 = load( symbolFile );

numSymbols  = length( x0 );
numPoints   = oversampling*numSymbols;

y0 = y0(1:numPoints);           % Ensure that y0 has the correct length
y0 = y0(:).';                   % Ensure that y0 and x0 are row vectors
x0 = x0(:).';

y0 = y0-mean( y0 );             % Remove any DC offset from the samples
x0 = x0/(max( x0 )-min( x0 ));  % Ensure that the symbols are bipolar and that
x0 = x0-1/2;                    % the minimum distance is 1

e1 = eye( eqNd+1, 1 );          % Define the identity vector, to be used later

%% Compute the frequency response of the anti aliasing filter %%%%%%%%%%%%%%%%%%
%% The anti-aliasing filter is a fourth-order Butterworth filter with the -3 dB
%% frequency set to 3/4 of the signaling speed.
freq = (-numPoints/2:numPoints/2-1)/numSymbols;

%% Denominator and numerator polynomials for the frequency response
ar = [1, 12.3141, 75.8181, 273.4537, 493.1335];
br = 493.1335;

Hr = br./polyval( ar, j*2*pi*freq );

%% Compute the frequency response for the spectral line bandpass filter %%%%%%%%
w1 = 2*pi*(1-1/1667);                       % Define the pass band
w2 = 2*pi*(1+1/1667);

%% Denominator and numerator polynomials for the bandpass filter
ap = [1, w2-w1, w1*w2];
bp = [0, w2-w1, 0];

Hp = polyval( bp, j*2*pi*freq )./polyval( ap, j*2*pi*freq );

%% Compute the equalized eye at the slicer input %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if any( graphOut )
    figNum = gcf;
end

for ii = 1:size( h0, 2 )        % stressor index
    %% Compute the stressor transfer function
    Hc = fftshift( fft( h0(:, ii), numPoints ) );
    Hc = Hc(:).';               % Ensure Hc is a row vector
    
    %% Process the waveform through the stressor and the anti-aliasing filter
    yr = real( ifft( fft( y0 ).*fftshift( Hc.*Hr ) ) );
    
    %% Compute the noise auto-correlation sequence %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% The noise at the anti-aliasing filter input is assumed to be white with
    %% power spectral density that yields the prescribed noise amplitude at the
    %% filter output. The output is then sampled at T/2 intervals.
    Rnn = real( ifft( abs( fftshift( Hr ) ).^2 ) );
    Rnn = (rmsNoise(ii).^2)*Rnn/Rnn(1);
    
    sampledRnn = Rnn(1:oversampling/2:end);
    sampledRnn = toeplitz( sampledRnn(1:eqNc) );
    
    %% Compute the sampling function and sample the processed waveform %%%%%%%%%
    km1 = mod( (0:numPoints-1)-1, numPoints )+1;
    kp1 = mod( (0:numPoints-1)+1, numPoints )+1;
    
    yclk = real( ifft( fft( (yr(kp1)-yr(km1)).^2 ).*fftshift( Hp ) ) );
    yclk = yclk(kp1)-yclk(km1);
    
    time = (0:numPoints)/oversampling;  % Wrap waveforms to ensure all edges are
    yr = [yr, yr(1)];                   % detected
    yclk = [yclk, yclk(1)];
    
    kr = find( diff( yclk > 0 ) > 0 );      % Eye center index
    kf = find( diff( yclk > 0 ) < 0 );      % Eye crossing index
    k = sort( [kr, kf] );
    index1 = double( kr(1) > kf(1) )+1;     % Index of the first eye center
    
    tk = time(k)-(1/oversampling)*yclk(k)./(yclk(k+1)-yclk(k));
    
    %% Compute the MMSE-DFE coefficients %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    yk = interp1( time, yr, mod( tk, time(end) ) );
    Y = toeplitz( yk, [yk(1), yk(end:-1:end-eqNc+2)] );
    Y = Y(index1:2:end, :);
    Ryy = Y'*Y+numSymbols*sampledRnn;

    %% Minimize MSE over the specified equalizer delay range %%%%%%%%%%%%%%%%%%%
    eqDelayList = (0:maxEqDelay-1);
    
    for jj = 1:length( eqDelayList )
        eqDelay = eqDelayList(jj);
        
        xd = x0(mod( (0:numSymbols-1)-eqDelay, numSymbols )+1);
        X = toeplitz( xd, [xd(1), xd(end:-1:end-eqNd+1)] );
        Rxx = X'*X;
        Ryx = Y'*X;
        
        lambda(jj) = 1/(e1'*inv( Rxx-Ryx'*inv( Ryy )*Ryx )*e1);
        D(:, jj) = lambda(jj)*inv( Rxx-Ryx'*inv( Ryy )*Ryx )*e1;
        C(:, jj) = inv( Ryy )*Ryx*D(:, jj);
    end     % for jj = length( eqDelayList )
    
    %% Install the best equalizer delay %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    [bestMSE, bestIndex] = min( lambda );
    bestEqDelay = eqDelayList(bestIndex);
    
    xd = x0(mod( (0:numSymbols-1)-bestEqDelay, numSymbols )+1);
    X = toeplitz( xd, [xd(1), xd(end:-1:end-eqNd+1)] );
    
    bestC = C(:, bestIndex);
    bestD = D(2:end, bestIndex);
    
    noiseVar = bestC'*sampledRnn*bestC;
    
    %% Generate slicer input eye diagram %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    phiList = linspace( -0.5, 0.5, round( 1/dphi )+1 );  
    veeList = linspace( -0.5, 0.5, round( 1/dvee )+1 );
        
    for jj = 1:length( phiList )
        phi = phiList(jj);
        
        yk = interp1( time, yr, mod( tk+phi, time(end) ) ); 
        Y = toeplitz( yk, [yk(1), yk(end:-1:end-eqNc+2)] );
        Y = Y(index1:2:end, :);
        
        zk = Y*bestC-X*[0; bestD];
        
        %% Compute the minimum distance from the noiseless, equalized samples
        %% to the decision threshold
        eyeLid0(jj) = max( zk(find( xd < 0 )) );
        eyeLid1(jj) = min( zk(find( xd > 0 )) );
        
        %% Compute the bit error ratio as a function of offset from the nominal
        %% sampling time and decision threshold
        dk = ones( length( veeList ), 1 )*zk.'-veeList(:)*ones( 1, numSymbols );
        dk(:, find( xd < 0 )) = -dk(:, find( xd < 0 ));
            
        berMap(:, jj) = mean( erfc( dk/sqrt( 2*noiseVar ) )/2, 2 );
    end     % for jj = 1:length( phiList )

    %% Normalize the coefficients so that the DC gain of the equalizer is 1 and
    %% compute the effective target amplitude, gamma
    gamma2(ii) = 1/sum( bestC );
    Cn(:, ii) = gamma2(ii)*bestC;
    Dn(:, ii) = gamma2(ii)*bestD;

    %% Compute the horizontal and vertical eye opening at the slicer input in 
    %% the absense of noise
    eyeLid = min( [-eyeLid0; eyeLid1] );
    
    kISI = find( abs( diff( eyeLid > 0 ) ) > 0 );
    phiISI = phiList(kISI)-dphi*eyeLid(kISI)./(eyeLid(kISI+1)-eyeLid(kISI));
    
    if length( phiISI ) == 0
        phiISI = [0, 0];
    end
    
    if length( phiISI ) == 1
        phiISI = sort( [phiISI, -sign( phiISI )/2] );
    end
    
    heyeISI(ii) = 2*max( min( [-phiISI(1), phiISI(2)] ), 0 );
    veyeISI(ii) = 2*gamma2(ii)*max( eyeLid(find( phiList == 0 )), 0 );
    
    %% Also compute the signal-to-noise ratio for reference
    snr(ii) = 10*log10( (veyeISI(ii)/gamma2(ii))^2/noiseVar );
    
    %% Compute the horizontal and vertical eye opening at the slicer input at
    %% the specified target bit error ratio
    phiTub = berMap(find( veeList == 0 ), :);
    veeTub = berMap(:, find( phiList == 0 )).';
    
    kBER = find( abs( diff( phiTub > ber0 ) ) > 0 );
    nBER = find( abs( diff( veeTub > ber0 ) ) > 0 );
    
    phiBER = phiList(kBER)-dphi*phiTub(kBER)./(phiTub(kBER+1)-phiTub(kBER));
    veeBER = veeList(nBER)-dvee*veeTub(nBER)./(veeTub(nBER+1)-veeTub(nBER));

    if (length( phiBER ) == 0) | (length( veeBER ) == 0)
        phiBER = [0, 0];
        veeBER = [0, 0];
    end
    
    if length( phiBER ) == 1
        phiBER = sort( [phiBER, -sign( phiBER )/2] );
    end
    
    heyeBER(ii) = 2*max( min( [-phiBER(1), phiBER(2)] ), 0 );
    veyeBER(ii) = 2*gamma2(ii)*max( min( [-veeBER(1), veeBER(2)] ), 0 );
    
    %% Present graphical output(s) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    if any( graphOut == 1 )     % MMSE-DFE slicer input eye diagram
        figure( figNum );
        clf;
               
        imagesc( phiList, gamma2(ii)*veeList, log10( berMap ) );
        
        hold on
        plot( phiList, gamma2(ii)*eyeLid0, '--', 'Color', 'white' );
        plot( phiList, gamma2(ii)*eyeLid1, '--', 'Color', 'white' );
        hold off
        
        jetColors = jet;
        colormap( jet );
        caxis( [round( log10( ber0 ) ), 0] );
        colorbar;
        set( gca, 'YDir', 'normal' );
        set( gca, 'Color', jetColors(end, :) );
        
        tapStr = [];
        tapStr = [tapStr, sprintf( '\nC = [%.3f', Cn(1, ii) )];
        for jj = 2:eqNc
            tapStr = [tapStr, sprintf( ', %.3f', Cn(jj, ii) )];
        end
        tapStr = [tapStr, ']'];
        if eqNd > 0
            tapStr = [tapStr, sprintf( '\nD = [%.3f', Dn(1, ii) )];
            for jj = 2:eqNd
                tapStr = [tapStr, sprintf( ', %.3f', Dn(jj, ii) )];
            end
            tapStr = [tapStr, ']'];
        else
            tapStr = [tapStr, sprintf( '\nD = []' )];
        end
        tapStr = [tapStr, sprintf( '\n\\gamma = ' )];
        tapStr = [tapStr, sprintf( '%.3f', gamma2(ii)/2 )];         
        
        eyeStr = [];
        eyeStr = [eyeStr, sprintf( 'SNR = %.1f dB\n', snr(ii) )];
        eyeStr = [eyeStr, sprintf( 'HEYE(ISI, 1E%.0f) = ', log10( ber0 ) )];
        eyeStr = [eyeStr, sprintf( '(%.3f, ', heyeISI(ii) )];
        eyeStr = [eyeStr, sprintf( '%.3f)\n', heyeBER(ii) )];
        eyeStr = [eyeStr, sprintf( 'VEYE(ISI, 1E%.0f) = ', log10( ber0 ) )];        
        eyeStr = [eyeStr, sprintf( '(%.3f, ', veyeISI(ii) )];
        eyeStr = [eyeStr, sprintf( '%.3f)\n', veyeBER(ii) )];
        
        text( -0.45,  gamma2(ii)*0.40, tapStr, 'Color', 'white' );
        text( -0.45, -gamma2(ii)*0.40, eyeStr, 'Color', 'white' );        
        title( sprintf( 'Bit error ratio map (%d)', ii ) );
        ylabel( 'Distance from decision threshold' );
        xlabel( 'Time (UI)' );
        
        figNum = figNum+1;
    end
    
end         % for ii = 1:size( h0, 2 )

warning on              % Restore warnings

%% End of script %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%